//===========================================================
//=     University of Illinois at Urbana/Champaign          =
//=     Department of Computer Science                      =
//=     Dr. Dan Roth - Cognitive Computation Group          =
//=                                                         =
//=  Project: Snow                                          =
//=                                                         =
//=   Module: Perceptron.cpp                                =
//=  Version: 2.1                                           =
//=  Authors: Jeff Rosen, Andrew Carlson                    =
//=     Date: xx/xx/99                                      = 
//=                                                         =
//= Comments:                                               =
//===========================================================

#include "Perceptron.h"
#include "Target.h"
#include "GlobalParams.h"
#include <math.h>
#include <iomanip>

// Presents an example to the Targets classifier.  The return
// value indicates whether the Targets classifier made a mistake
// on this example (and had its weight vector updated).
// Return of true indicates a mistake and update, otherwise false.

bool Perceptron::PresentExample(  Target& tar, Example& ex )
{
    // Reset activation flags
    tar.externalActivation = ex.HasFeature(tar.targetID);
    tar.internalActivation = 0.0;
    
    Example::FeatureMap::const_iterator it = ex.GetFeatureBegin();
    Example::FeatureMap::const_iterator end = ex.GetFeatureEnd();

    Target::FeatureMap::iterator f;
    
    double totalWeight = 0.0;
    
    for ( ; it != end; it++)
    {
        // Don't include this target
        if (it->first != tar.targetID) 
        {
            // Does the feature already exist?
            if (((f = tar.features.find(it->first)) != tar.features.end()) &&
                (f->second.discarded != pending))
            {
                if (tar.externalActivation) 
                {
                   if (globalParams.currentCycle == 1)
                   {
                      f->second.activeCount++;
                   }
                }
                if (f->second.discarded == active)
                {
                    totalWeight += it->second * f->second.weight;
                }
            } else {
                // either feature doesn't exist or is pending
                int actualCount = 1;

                // Get the actual global feature count
                if (globalParams.eligibilityThreshold > 1)
                {
                    Target::FeatureMap::iterator fc;
                    if ((fc = tar.features.find(it->first))
                        != tar.features.end())
                    {   
                        if (tar.externalActivation
                            && (globalParams.currentCycle == 1)) 
                        {
                            fc->second.activeCount++;
                        }
                        actualCount = fc->second.activeCount;
                    } else {
                        // Add the feature
                        tar.features.insert(
                            make_pair(it->first, MinimalFeature()));
                        tar.features[it->first].discarded = pending;
                        if (tar.externalActivation) 
                        {
                            tar.features[it->first].activeCount = 1;
                        }
                    }
                }

                // Check the eligibility of the feature
                if ((actualCount >= globalParams.eligibilityThreshold)
                    && ((globalParams.sparseNetwork && tar.externalActivation)
                        || !globalParams.sparseNetwork))
                {
                    if (globalParams.verbosity >= VERBOSE_MAX)
                        cout << "Perceptron adding feature (" << it->first
                             << ") to target (" << tar.targetID
                             << ")" << endl;
                
                    // If it is eligible, set its weight
                    // (and increment the totalWeight)
                    if (tar.externalActivation)
                    {
                        tar.features[it->first].discarded = active;
                        tar.features[it->first].activeCount++;
                   }
                    
                    tar.features[it->first].weight = defaultWeight;
                    //totalWeight += defaultWeight;
                }
            }
        } 
    }

    // Update the active / non-active counters
    if (tar.externalActivation) tar.activeCount++; else tar.nonActiveCount++;

    // Update internal activation, if necessary
    tar.internalActivation = totalWeight; 
    bool intActivation = (tar.internalActivation >= threshold);
    
    // If there is a mistake let the algorithm update the weights
    if ((tar.externalActivation != intActivation)
        && ((globalParams.currentCycle > 1)
        || !globalParams.noFirstCycleUpdate))
    {
        if (globalParams.verbosity >= VERBOSE_MED)
            cout << "Updating target (" << tar.targetID
                 << ") for mistake..." << endl;
        
        Update(tar, ex);

        return true;
    }

    return false;
}


// Updates the target based on the example and the Perceptron parameters

void Perceptron::Update( Target& tar, Example& ex )
{
    // Assume false negative
    double delta = learningRate;

    // Update only called if a mistake was made.
    // If the target was internally activated, then the mistake
    // must have been a false positive.
    if (tar.internalActivation >= threshold) delta *= -1.0;

    // Declare the iterators we will use to access the intersection
    // of the examples active features with the targets weight
    // vector.
    Example::FeatureMap::const_iterator it = ex.GetFeatureBegin();
    Example::FeatureMap::const_iterator end = ex.GetFeatureEnd();

    Target::FeatureMap::iterator f;

    // Update the weight vector
    for ( ; it != end; it++)
    {
        if ((f = tar.features.find(it->first)) != tar.features.end())
        {
           if (tar.features[it->first].discarded == active)
           {
            f->second.weight += delta;
            f->second.updates++;
           }
        }
    }

    // Update the confidence of the target
    double x = sqrt((double)(tar.activeCount + tar.nonActiveCount));
    tar.confidence /= x / (((x - 1) / (delta + 1)) + 1);
}

void Perceptron::SetTargetActivation( Target& tar, Example& ex )
{
    Example::FeatureMap::const_iterator it = ex.GetFeatureBegin();
    Example::FeatureMap::const_iterator end = ex.GetFeatureEnd();

    Target::FeatureMap::iterator f;
    
    tar.externalActivation = ex.HasFeature(tar.targetID);
    if (tar.externalActivation) tar.activeCount++; else tar.nonActiveCount++;
    tar.internalActivation = 0.0;
    
    for ( ; it != end; it++)
    {
        // Don't process the target itself
        if (it->first != tar.targetID)
        {
            // Does this feature have a weight and is it active?
            if (((f = tar.features.find(it->first)) != tar.features.end()) 
                && (f->second.discarded == active) )
            {
                tar.internalActivation += it->second * f->second.weight;
            } else {
                // Don't smooth any targets!
                if (globalParams.targetIds.find(it->first) ==
                                  globalParams.targetIds.end())
                    tar.internalActivation -= it->second * globalParams.smoothing;
            }
        }
    }

}

double Perceptron::ReturnNormalizedActivation( Target& tar)
{
    double normalizedActivation = 
       (1.0 / (1.0 + exp(-(tar.InternalActivation() -
                                  threshold) )) ) * tar.Confidence();
    return normalizedActivation;
}

void Perceptron::Show( ostream& out )
{
    out << "Perceptron network:("
        << setprecision(4) << defaultWeight << ", "
        << setprecision(4) << threshold << ", "
        << setprecision(4) << learningRate << ")" << endl;
}

void Perceptron::Read( ifstream& in )
{
    in >> learningRate >> threshold >> defaultWeight;
    
    if (globalParams.verbosity >= VERBOSE_MED)
    {
        cout << endl << "Created:";
        Show(cout);
    }
}

void Perceptron::Write( ofstream& out )
{
    if (globalParams.verbosity >= VERBOSE_MED)
    {
        Show(cout);
    }

#ifdef WIN32
    // This looks weird, but it is a work around for the
    // NT version.  Due to a bug in MSVC 5.0 you have to
    // call the insertion operator explicitly.
    operator<<(out, "perceptron");
    
    out << ' '
        << index << ' '
        << learningRate << ' '
        << threshold << ' '
        << defaultWeight << endl;
#else
    out << "perceptron "
        << index << ' '
        << learningRate << ' '
        << threshold << ' '
        << defaultWeight << endl;
#endif
}

    


